/*
 *  GenericViewer.cc
 *
 *  Generic Viewer class derived from Viewer
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "GenericViewer.h"

using SigC::bind;
using SigC::slot;

GenericViewer::GenericViewer(char* label):  
  Viewer(label),
  mouse_mode(1)  // this button is toggled first in the constructor of camera_control!!!
{
  
  /* initialize camera_control */
  /* connecting child -> this update signals */
  Viewer::gl_changed_event.connect(slot(this,&GenericViewer::update_camera_control));
  camera_control.camera_Gtk_CoordEntry.entry_activated.connect(slot(this,&GenericViewer::entry_activated));
  camera_control.lookat_Gtk_CoordEntry.entry_activated.connect(slot(this,&GenericViewer::entry_activated));
  camera_control.up_Gtk_CoordEntry.entry_activated.connect(slot(this,&GenericViewer::entry_activated));
  camera_control.light_Gtk_CoordEntry.entry_activated.connect(slot(this,&GenericViewer::entry_activated));

  /* connecting mode change signal */
  camera_control.mode_set.connect(slot(this,&GenericViewer::set_mouse_mode));

  /* connecting button signals */
  camera_control.main_button.clicked.connect( bind<int>(slot(this,&GenericViewer::tool_handler),1));

  /* connecting this -> child update signals */
  camera.pos_update.connect(slot(&camera_control.camera_Gtk_CoordEntry,&CoordEntry::set_coord));
  camera.lookat_update.connect(slot(&camera_control.lookat_Gtk_CoordEntry,&CoordEntry::set_coord));
  camera.up_update.connect(slot(&camera_control.up_Gtk_CoordEntry,&CoordEntry::set_coord));
  
  /* setting window in front of Viewer */
  camera_control.set_transient_for(*this);

  /* initialize CoordEntries */
  camera_control.camera_Gtk_CoordEntry.set_coord(camera.pos);
  camera_control.lookat_Gtk_CoordEntry.set_coord(camera.lookat);
  camera_control.up_Gtk_CoordEntry.set_coord(camera.up);
  camera_control.light_Gtk_CoordEntry.set_coord(gl_light_pos);

  
  /* initialize light_control */
  /* connecting child -> this update signals */
  light_control.light_set.connect(slot( this, &GenericViewer::set_light));

  /* connecting button signals */
  light_control.main_button.clicked.connect( bind<int>(slot(this,&GenericViewer::tool_handler),2));
  
  /* setting window in front of Viewer */
  light_control.set_transient_for(*this);

  
  /* setting up the toolbar */

  using namespace Gtk::Toolbar_Helpers;
  Viewer::upper_hbox.pack_start(toolbar);
  toolbar.tools().push_back(Space());
  toolbar.tools().push_back(Space()); 

  /* these are the generic viewer specific tools...*/
  gv_toolbar.tools().push_back(Element(camera_control.main_button,"edit camera stats"));
  gv_toolbar.tools().push_back(Element(light_control.main_button,"change lighting"));
  
  /* ...packed together into a frame..*/
  gv_toolbar_frame.set_border_width(10);
  gv_toolbar_frame.set_shadow_type(GTK_SHADOW_ETCHED_IN);
  gv_toolbar_frame.set_label(" Viewer Tools");
  gv_toolbar_frame.add(gv_toolbar);
  
  /*... which is inserted into the main toolbar */
  toolbar.tools().push_back(Element(gv_toolbar_frame));

}


 
gint GenericViewer::gl_motion_notify_handler_imp(GdkEventMotion* event)
{
  /* old mode so the Viewer methods are called */ 
  if (mouse_mode == 1) 
    Viewer::gl_motion_notify_handler_imp(event);
  /* new mode */  
  else if (mouse_mode == 2) 
    {

      gint x,y;
      
      GdkModifierType state;
      
      if (event->is_hint)
	gdk_window_get_pointer (event->window, &x, &y, &state);
      else
	{
	  x = (gint) event->x;
	  y = (gint) event->y;
	  state =(GdkModifierType) event->state;
	}
      
      if ((state & GDK_BUTTON1_MASK)&&mouse1_pressed)
	{
	  Vector3 look=Viewer::camera.lookat-Viewer::camera.pos;
	  look=look/look.abs();
	  Vector3 up=Viewer::camera.up;
	  Vector3 side=up^look;
	  Vector3 change=up*(mouse_start_y-y)/100
	                 +side*(mouse_start_x-x)/100;
	  mouse_start_y=y;
	  mouse_start_x=x;

	  Viewer::camera.lookat=Viewer::camera.lookat+change;
	  Viewer::camera.pos=Viewer::camera.pos+change;
 	 
	  Viewer::updategl();
	  Viewer::drawgl();


	  camera_control.camera_Gtk_CoordEntry.set_coord(camera.pos);
	  camera_control.lookat_Gtk_CoordEntry.set_coord(camera.lookat);
	  camera_control.up_Gtk_CoordEntry.set_coord(camera.up);

	}
      if ((state & GDK_BUTTON2_MASK)&&mouse2_pressed)
	{
	  Vector3 look=Viewer::camera.lookat-Viewer::camera.pos;
	  look=look/look.abs();
	  
	  Vector3 change=look*(mouse_start_y-y)/100;
	  mouse_start_y=y;
	  mouse_start_x=x;
	 
	  if (mouse_start_y-y>0)
	    {
	      Viewer::camera.lookat=Viewer::camera.lookat+change;
	      Viewer::camera.pos=Viewer::camera.pos+change;
	    }
	  else
	    {
	      Viewer::camera.pos=Viewer::camera.pos+change;
	      Viewer::camera.lookat=Viewer::camera.lookat+change;
	    }


	  Viewer::updategl();
	  Viewer::drawgl();

	  
	  camera_control.camera_Gtk_CoordEntry.set_coord(camera.pos);
	  camera_control.lookat_Gtk_CoordEntry.set_coord(camera.lookat);
	  camera_control.up_Gtk_CoordEntry.set_coord(camera.up);
	}
      
      
      if ((state & GDK_BUTTON3_MASK)&&mouse3_pressed)
	{


	}
      
    }
}

gint GenericViewer::gl_button_press_handler_imp(GdkEventButton* event)
{
  /* old mode so the Viewer methods are called */ 
  if (mouse_mode == 1) 
    Viewer::gl_button_press_handler_imp(event);
  
  /* new mode */    
  else if(mouse_mode == 2){
    if (event->button==1){
      mouse_start_x=event->x;
      mouse_start_y=event->y;
      mouse1_pressed=1;
    }
    if (event->button==2){
      mouse_start_x=event->x;
      mouse_start_y=event->y;
      mouse2_pressed=1;
    }
    
    if (event->button==3){
      mouse_start_x=event->x;
      mouse_start_y=event->y;
      mouse3_pressed=1;
  }
  }
}

gint GenericViewer::gl_button_release_handler_imp(GdkEventButton* event)
{ 
  /* old mode so the Viewer methods are called */ 
  if (mouse_mode == 1) Viewer::gl_button_release_handler_imp(event);

  /* new mode */  
  else if (mouse_mode == 2)
    {
      if (event->button==1)  
	{
	  mouse1_pressed=0;
	}
      
      if (event->button==2)  
	{
	  mouse2_pressed=0;
	}
      
      if (event->button==3)  
	{
	  mouse3_pressed=0;
	}
      
    }
}


void GenericViewer::set_light()
{
  Viewer::set_light(light_control.get_light());
  updategl();
  drawgl();
}



void GenericViewer::update_camera_control()
{
  camera_control.camera_Gtk_CoordEntry.set_coord(camera.pos);
  camera_control.lookat_Gtk_CoordEntry.set_coord(camera.lookat);
  camera_control.up_Gtk_CoordEntry.set_coord(camera.up);
  camera_control.light_Gtk_CoordEntry.set_coord(gl_light_pos);
}



void GenericViewer::tool_handler(int id)
{

  switch(id) {

  /* the CameraControl window will be showed/hidden when its button is pressed */
  case 1:
    if(camera_control.active == 0)
      {
	camera_control.set_active(true);
      }
    else
      {
	camera_control.set_active(false);
      }
    break;

  /* as well as the LightControl window */
  case 2:     
    if(light_control.active == 0)
      {
	light_control.set_active(true);
      }
    else
      {
	light_control.set_active(false);
      }
    break;
  }

}



void GenericViewer::entry_activated(int id){

  //need to update up first as it might be changed later due to orthogonalization
  //   if (id==camera_control.up_Gtk_CoordEntry.get_id())
  Viewer::camera.set_up(camera_control.up_Gtk_CoordEntry.get_coord());
     //  if (id==camera_control.camera_Gtk_CoordEntry.get_id())
  Viewer::camera.set_pos(camera_control.camera_Gtk_CoordEntry.get_coord());
   
     //  if (id==camera_control.lookat_Gtk_CoordEntry.get_id())
  Viewer::camera.set_lookat(camera_control.lookat_Gtk_CoordEntry.get_coord());
   
     //  if (id==camera_control.light_Gtk_CoordEntry.get_id())
  Viewer::set_light(camera_control.light_Gtk_CoordEntry.get_coord());
  
  updategl();
  drawgl();
}
