/*
 *  QuatFractCreator.h
 *
 *  The center of your fractal world
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef QUAT_FRACT_H
#define QUAT_FRACT_H

#include "QuatMath.h"
#include "Cube.h"
#include <list>
#include <vector>
#include <map>
#include <string>
//only for now...
#include <gtk--/frame.h>

class QuatFractCreator;



typedef enum { CUBE, TETRA, COMBI, TRACK} pm;
typedef enum { SQR, CUBIC} iter_func;

/** Fractal parameters class.
    This class acts as a container of some options, parameters concerning
    the fractal drawing.
*/
class FractParams{
  
 public:

  /** which method is used for polygonization */
  pm polyMethod;

  /** the type of function to be iterated */
  iter_func iterFunc;

  /** drawing box lower boundary of x coord */
  double xMin;
  /** drawing box lower boundary of y coord */
  double yMin;
  /** drawing box lower boundary of z coord */
  double zMin;
  /** drawing box upper boundary of x coord */
  double xMax;
  /** drawing box upper boundary of y coord */
  double yMax;
  /** drawing box upper boundary of z coord */
  double zMax;
  
  /** step size used for rendering the julia set */
  double step;

  /** maximal iterations done to test convergence */
  int maxiter;

  /** if the abs is greater than bailout after maxiter iterations
      the sequence is supposed to diverge */
  double bailout;

  /** the quaternion number in the iteration function f(q) = q^2 + C */
  Quat C;

  /** x4 coordinate. The slice with the 4-dim space is done at x4 */
  double x4;

};


/** Quaternion Fractal Creator class.
    This class provides the interface to the Creation of Julia sets of
    Quaternion Fractals. It is right now specialized for use with the
    Cube Polygonization Algorithm. 
*/
class QuatFractCreator{

  /** Fractal and rendering parameters */
  FractParams fparams;

  /** Cube poly cache. Mapping every cube variant to its polygon.*/
  // TODO this should be static...
  vector< list<unsigned char>* > poly_map;

  /** Cube normal cache. Mapping every cube variant to its normal.*/
  vector< vector<double>* > normal_map;

  /** Convergence map. Mapping the quaternion points in the rendering
      area to IN or OUT (meaning in or out of the prisoner set).*/
  vector< unsigned char > conv_map;

  /** Cube vector. Saves all cubes containing surface polygons.*/
  vector< Cube> cube_v;

  /* map the surface vertices to the list of 
     the (four) polygons it is involved in */
  map< vector<double>, list<int> > point_to_polys;

  /* map the surface vertices to their normals 
     , this saved redundant calculations */
  map< vector<double>, vector<double> > point_to_normal;

 public:
  
  /** Signal to update progress bar */
  SigC::Signal1 <void, float> render_progress;

  /** Constructor. A Quaternion Fractal with C = [a b c d] is created. */
  QuatFractCreator( double a, double b, double c, double d);
  
  /** Set/Get rendering method. By now only Cube rendering gives cool results but
      in future there should be several different rendering methods that
      one can choose from with this method.*/
  inline void set_method(pm m) { fparams.polyMethod = m; }
  inline pm get_method() { return fparams.polyMethod; }

  /** Set/Get iteration function. This gives access to the function which
      is iterated when creating the Julia set */
  void set_func(iter_func m);
  inline iter_func get_func() { return fparams.iterFunc; }

  /** Set the fractal parameters. Simply overwrites the actual fractal parameters
      saved in fparams with the given ones.*/
  void set_fparams(const FractParams& fp);

  /** Get the fractal parameters. Should be clear.*/
  inline FractParams get_fparams() { return this->fparams; }
  
  /** Convergence check. Tests if the actual sequence converges at q.*/
  bool iter_sqr(Quat &q);
  bool iter_cubic(Quat &q);

  bool (QuatFractCreator::* konv)(Quat& q);

  /* Initialize convergence cache. Iterates through the area that will be
     rendered (defined by fparams) and sets up the convergence map.*/
  // void init_conv_map(); // this is now done on the fly !

  /** Tetra polygonizing algorithm. Polygonize the fractal using tetraeders
      to calculate the polygons.*/
  void tetra_polygonize();

  /** Tetra test method. Do some debugging tests on tetra polygonizer.*/
  void tetra_test();

  /** Initialize the cube cache. For all 254 variants of cube edge configurations
      the polygons are defined by a list<char> containing indices to vectors
      pointing on the edges of the cube. The normals of each cube variant is
      cached, too.*/
  void cube_init(); 

  /** Cube polygonizing algorithm. Polygonize the fractal using cubes. */
  void cube_polygonize();

  /** Merge normals and draw fractal. Merge the normals of neighbor cubes
      and draw the polygons in the cubes.*/
  void cube_merge_n_draw();

  /** Cube polygonizing tests. */
  void cube_test();

  /** Vrend output. */
  void write_vrend(char* outfile);
  
  /** Surface tracking algorithm. It is not implemented yet */
  void track_polygonize();

  //Ignore this
  void register_data (SigC::ScopeNode *);
};



#endif
