/*
 *  SplineMath.cc
 *
 *  Mathematical interface to spline interpolation and rendering
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */



#include "SplineMath.h"


SplineMath::SplineMath() : render(false) {};


SplineMath::SplineMath(  const vector<Waypoint*>& pp) :
  path(pp), render(false) {};


void SplineMath::get_splines() {

  /* we will need n spline parts for each dimension... */
  int n = path.size() - 1;
  /* ... and store them in here */
  vector< SplinePart* >* spp = new vector< SplinePart* >;

  /* initialize Spline vector */
  int spp_c = 0;
  for( vector< Waypoint*>::iterator wi = path.begin();
       wi != path.end()-1; wi++) {
    spp->push_back( new SplinePart(10, (*wi)->time));
    spp_c++;
  }
  spp->push_back( new SplinePart(10,  path.back()->time + 0.01));


  /* get 1-dimensional spline coefficients for each dimension */
  // we have 10 dimensions (pos, lookat, up and x4)
  for( int dim = 0; dim < 10; dim++) {

    vector<double> t(n+1);
    vector<double> f(n+1);
    vector<double> mue(n+1), rho(n+1), q(n+1), lam(n+1), del(n+1), 
      a(n+1), b(n+1), c(n+1), d(n+1), y(n+1);


    /* initialize sampling points */
    int j = 0;
    for( vector< Waypoint*>::iterator wi = path.begin(); 
	 wi != path.end(); wi++) {
      t[j] = (*wi)->time;
      if( dim < 3)
	f[j] = (*wi)->pos[dim];
      else if( dim < 6)
	f[j] = (*wi)->lookat[dim-3];
      else if( dim < 9)
	f[j] = (*wi)->up[dim-6];
      else
	f[j] = (*wi)->x4;
      j++;
    }


    /* the spline polynons have the form 
       a_i + b_i*(t-t_i) + c_i/2*(t-t_i)^2 + d_i/6*(t-t_i)^3
       we will first get the c_i :*/

    /* some helper variables */
    for( int i = 1; i < n; i++) {
      del[i] = 6*( (f[i+1] - f[i])/(t[i+1] - t[i]) - (f[i] - f[i-1])/(t[i] - t[i-1]))
	/ ( t[i+1] - t[i-1]);
      mue[i] = (t[i+1] - t[i])/(t[i+1] - t[i-1]);
      lam[i] = (t[i] - t[i-1])/(t[i+1] - t[i-1]);
    }

    /* LR decomposition */
    q[1] = mue[1] / 2;
    for( int i = 2; i < n-1; i++) {
      rho[i] = 2 - q[i-1]*mue[i];
      q[i] = lam[i] / rho[i];
    }
    rho[n-1] = 2 - q[n-2]*mue[n-1];
  
 
    /* solve Ly = f */
    y[1] = del[1] / 2;
    for( int i = 2; i < n; i++) 
      y[i] = (del[i] - mue[i]*y[i-1]) / rho[i];
  

    /* solve Rx = y */
    c[0] = c[n] = 0;
    c[n-1] = y[n-1];
    for( int i = n-2; i > 0; i--) 
      c[i] = y[i] - q[i]*c[i+1];

    /* now we easily can get a_i, b_i, d_i from the c_i :*/
    for( int i = 0; i < n; i++) {
      ((*spp)[i])->pc_vec[dim].a = f[i];
      ((*spp)[i])->pc_vec[dim].b = (f[i+1] - f[i])/(t[i+1] - t[i]) - 
	(c[i+1] + 2*c[i])*(t[i+1] - t[i])/6;
      ((*spp)[i])->pc_vec[dim].c = c[i];
      ((*spp)[i])->pc_vec[dim].d = (c[i+1] - c[i]) / (t[i+1] - t[i]);

    }

  }
  sp_part = spp;
  render = true;
}


void SplineMath::print_splines() {  
  
  for( vector< SplinePart*>::iterator si = sp_part->begin(); 
       si != sp_part->end(); si++) {
    cout << "[ ";
    for( int dim = 0; dim < 10; dim++) {
      cout << "[ " << (*si)->pc_vec[dim].a << " " <<
	(*si)->pc_vec[dim].b << " " <<
	(*si)->pc_vec[dim].c << " " <<
	(*si)->pc_vec[dim].d << " ]" << endl;
    }
    cout << "]" << endl;
  }

}


vector< double> SplineMath::get_spline_at( double t) {
  
  vector< double> ret(3);
  vector< Waypoint*>::iterator wi = path.begin();
  int j = 0;
  
  /* which spline part should we take? */
  while( (t > (*(wi+1))->time) && (wi != path.end()-1)) {
    j++;
    wi++;
  }

  ret[0] = ((*sp_part)[j])->pc_vec[0].a + 
    ((*sp_part)[j])->pc_vec[0].b * (t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[0].c/2 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[0].d/6 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t);

  ret[1] = ((*sp_part)[j])->pc_vec[1].a + 
    ((*sp_part)[j])->pc_vec[1].b * (t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[1].c/2 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[1].d/6 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t);

  ret[2] = ((*sp_part)[j])->pc_vec[2].a + 
    ((*sp_part)[j])->pc_vec[2].b * (t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[2].c/2 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t) + 
    ((*sp_part)[j])->pc_vec[2].d/6 * 
    (t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t)*(t - ((*sp_part)[j])->t);

    
  return ret;

}  


void SplineMath::render_splines() {  

  if(render) {
    vector< double> a(3);
    
    glDisable(GL_LIGHTING);
    glColor3f(1, 1, 0);
    glBegin(GL_LINE_STRIP);
    
    for( double t = path.front()->time; t <= path.back()->time; t += 0.1) {
      a = get_spline_at( t);
      glVertex3f(a[0], a[1], a[2]);
    }
    
    glEnd();
    glEnable(GL_LIGHTING);
  }
     
}
  
