/*
 *  SplineMath.h
 *  
 *  Mathematical interface to spline interpolation and rendering
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#ifndef SPLINEMATH_H
#define SPLINEMATH_H

#include <stdio.h>
#include <vector>

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "Waypoint.h"

typedef enum { NATURAL, PERIODIC, CLAMPED} spline_type;

/* data representing a vector of 1-dimensional spline coefficients */
struct SplinePoly {
  SplinePoly() {};
  SplinePoly( double aa, double bb, double cc, double dd) :
    a(aa), b(bb), c(cc), d(dd) {};
  double a, b, c, d;
};

/* data representing a part of a n-dimensional spline (for time > t) */
struct SplinePart {
  SplinePart( int n) : pc_vec(n) {};
  SplinePart( int n, double tt) :
    t(tt), pc_vec(n){};
  vector< SplinePoly> pc_vec; //polynom coefficients
  double t; // start time
};


/** Spline Mathematics. Provides an interface to 
    natural spline interpolation */
class SplineMath {

  /** Render enable. Is there a spline that can be rendered? */
  bool render;

  /** Waypoint path. This path will be interpolated by the spline */
  vector< Waypoint*> path;

  /** SplinePart vector. The whole spline is stored here as a vector 
      of the spline parts ordered by their start_time */
  vector< SplinePart*>* sp_part;
  
public:
  /** Constructor. */
  SplineMath();
  /** Constructor. Accepting sampling point vector. */
  SplineMath( const vector<Waypoint*>& path); 
  
  /** Update waypoint path. Updates the path to be interpolated by splines. */
  void update_path( const vector<Waypoint*>& pp) { path = pp;} 

  /** Get splines vector. Solves and stores the splines coefficients. */
  void get_splines();

  /** Render splines. Creating a spline drawing with OpenGL calls. */
  void render_splines();

  /** Print the spline coefficients to stdout. debugging function */
  void print_splines();

  /** Get the spline coefficients vector. */
  vector< SplinePart*>* get_spline_data() { 
    if( render == true) 
      return sp_part;
    else 
      return NULL;
  }

  /** Get the spline function vector at time t. */
  vector< double> get_spline_at( double t);
  
};


#endif
