/*
 *  TetraCube.cc
 *
 *  Helper Class for Tetra Polygonization Algorithm
 *
 *  Copyright (C) 2002 Tobias Dyckerhoff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include "TetraCube.h"


/* Constructor */

TetraCube::TetraCube(int i) : v(i), in(i) {
  inmap = 0;
  in_count = 0;
}


/* return vertex between i and j */

vector<double> TetraCube::inbetw( int i, int j) {
  vector<double> a(3);
  a[0] = (v[i][0] + v[j][0]) /2;
  a[1] = (v[i][1] + v[j][1]) /2;
  a[2] = (v[i][2] + v[j][2]) /2;

  return a;
}



/* polygonizes the n-th tetraeder in the cube  */

list<vector<double> > TetraCube::tetra_polygonize( int n) {
  /* this is the array of polygon indices on which the slices operate */
  //TODO indirect array ?
  const int m[] = { 0, 3, 1, 5, 
		    0, 3, 7, 2, 
		    3, 0, 7, 5, 
		    4, 5, 7, 0,
		    0, 4, 2, 7, 
		    6, 4, 7, 2};

  char pol_map = 0x0;
  double sign = 1;

  /* get the bitmap encoding the triangle */
  for( int i = 0; i < 4; i++) 
    pol_map |= in[m[4*n+i]] << (3-i);
    

  list<vector<double> > tetra_polygon;

  if( pol_map != 0 && pol_map != 0xF) {
  
    /* we make use of symmetries so we just have to check 
       the last 3 bits of pol_map and get the orientation
       of the polygon from the last one */
    if( pol_map & 0x8) {
      sign = -1;
      // cout << "inverted" << endl;
      pol_map ^= 0xF;
    }
  
    /* save some vectors defining the normal */
    vector<double> a;
    vector<double> b;
    vector<double> c;
    switch( pol_map) {
    case 0x1:
      tetra_polygon.push_back( a = inbetw(m[4*n+0], m[4*n+3]));
      tetra_polygon.push_back( b = inbetw(m[4*n+1], m[4*n+3]));
      tetra_polygon.push_back( c = inbetw(m[4*n+2], m[4*n+3]));
      break;
    case 0x2:
      tetra_polygon.push_back( a = inbetw(m[4*n+3], m[4*n+2]));
      tetra_polygon.push_back( b = inbetw(m[4*n+1], m[4*n+2]));
      tetra_polygon.push_back( c = inbetw(m[4*n+0], m[4*n+2]));
      break;
    case 0x3:
      tetra_polygon.push_back( a = inbetw(m[4*n+1], m[4*n+3]));
      tetra_polygon.push_back( b = inbetw(m[4*n+1], m[4*n+2]));
      tetra_polygon.push_back( c = inbetw(m[4*n+0], m[4*n+2]));
      tetra_polygon.push_back( inbetw(m[4*n+0], m[4*n+3]));
      break;
    case 0x4:
      tetra_polygon.push_back( a = inbetw(m[4*n+3], m[4*n+1]));
      tetra_polygon.push_back( b = inbetw(m[4*n+0], m[4*n+1]));
      tetra_polygon.push_back( c = inbetw(m[4*n+2], m[4*n+1]));
      break;
    case 0x5:
      tetra_polygon.push_back( a = inbetw(m[4*n+0], m[4*n+3]));
      tetra_polygon.push_back( b = inbetw(m[4*n+0], m[4*n+1]));
      tetra_polygon.push_back( c = inbetw(m[4*n+1], m[4*n+2]));
      tetra_polygon.push_back( inbetw(m[4*n+2], m[4*n+3]));
      break;
    case 0x6:
      tetra_polygon.push_back( a = inbetw(m[4*n+3], m[4*n+1]));
      tetra_polygon.push_back( b = inbetw(m[4*n+0], m[4*n+1]));
      tetra_polygon.push_back( c = inbetw(m[4*n+0], m[4*n+2]));
      tetra_polygon.push_back( inbetw(m[4*n+3], m[4*n+2]));
      break;
    case 0x7:
      tetra_polygon.push_back( a = inbetw(m[4*n+0], m[4*n+3]));
      tetra_polygon.push_back( b = inbetw(m[4*n+0], m[4*n+1]));
      tetra_polygon.push_back( c = inbetw(m[4*n+0], m[4*n+2]));
      break;  
    default:
      cerr << "not possible: all tetras missed!!" << endl;
    }

    //TODO: cache this normal! */
    /* get the normal vector as a-b x c-b */
    vector<double> normal(3);
    normal[0] = (a[1] - b[1])*(c[2] - b[2]) - (a[2] - b[2])*(c[1] - b[1]);
    normal[1] = (a[2] - b[2])*(c[0] - b[0]) - (a[0] - b[0])*(c[2] - b[2]);
    normal[2] = (a[0] - b[0])*(c[1] - b[1]) - (a[1] - b[1])*(c[0] - b[0]);
  
    double nabs = sqrt(normal[0]*normal[0]
		       +normal[1]*normal[1]
		       +normal[2]*normal[2]);
    if( nabs == 0)
      cerr << "aaaaaaaaaaaaaaaaaahh no normal" << endl;
    for( int i = 0; i < 3; i++) {
      normal[i] /= sign*nabs;
    }

    tetra_polygon.push_back( normal);
  }
  return tetra_polygon;
}



/* this approach decomposes the cube into 6 triangles 
   so we get only polygons with 3 or 4 corners that induce
   ONE plane! -> normals are right */
list<list<vector<double> > > TetraCube::polygonize() {

  if( in_count > 0 && in_count < 8) {
    
    /* this list will store all tetra_polygons in the cube */
    list<list<vector<double> > > polygons;

    for( int i = 0; i < 8; i++) {
      inmap |= (in[i] << (7-i));
    }

    /* decompose the cube into triangles using slices */
    for( int i = 0; i < 6; i++) {
      list<vector<double> > tetra_polygon = tetra_polygonize( i);
      //cout << tetra_polygon.size();

      if( tetra_polygon.size() != 0) {
	/* push polygon into list */
	polygons.push_back( tetra_polygon);
      }
      
    }
    return polygons;
  }
}
